/***************************************************************************
 *                                                                         *
 *  MAXIMUS-CBCS Source Code, Version 1.02                                 *
 *  Copyright 1989, 1990 by Scott J. Dudley.  All rights reserved.         *
 *                                                                         *
 *  SCANBLD main module                                                    *
 *                                                                         *
 *  For complete details of the licensing restrictions, please refer to    *
 *  the licence agreement, which is published in its entirety in MAX.C     *
 *  and LICENCE.MAX.                                                       *
 *                                                                         *
 *  USE OF THIS FILE IS SUBJECT TO THE RESTRICTIONS CONTAINED IN THE       *
 *  MAXIMUS-CBCS LICENSING AGREEMENT.  IF YOU DO NOT FIND THE TEXT OF THIS *
 *  AGREEMENT IN ANY OF THE  AFOREMENTIONED FILES, OR IF YOU DO NOT HAVE   *
 *  THESE FILES, YOU SHOULD IMMEDIATELY CONTACT THE AUTHOR AT ONE OF THE   *
 *  ADDRESSES LISTED BELOW.  IN NO EVENT SHOULD YOU PROCEED TO USE THIS    *
 *  FILE WITHOUT HAVING ACCEPTED THE TERMS OF THE MAXIMUS-CBCS LICENSING   *
 *  AGREEMENT, OR SUCH OTHER AGREEMENT AS YOU ARE ABLE TO REACH WITH THE   *
 *  AUTHOR.                                                                *
 *                                                                         *
 *  You can contact the author at one of the address listed below:         *
 *                                                                         *
 *  Scott Dudley           FidoNet  1:249/106                              *
 *  777 Downing St.        IMEXnet  89:483/202                             *
 *  Kingston, Ont.         Internet f106.n249.z1.fidonet.org               *
 *  Canada - K7M 5N3       BBS      (613) 389-8315 - HST/14.4K, 24hrs      *
 *                                                                         *
 ***************************************************************************/

#define NOVARS
#define MAX_SCANAREA 128
#define HASH_SIZE int

#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dos.h>
#include <string.h>
#include <ctype.h>
#include <mem.h>
#include "prog.h"
#include "ffind.h"
#include "max.h"
#include "dr.h"

#ifdef OS_2
#undef  MAX_MSG
#define MAX_MSG 4000
#endif

char orig_disk;       /* Disk drive number (0=A, 1=B, etc) we started from */
char *orig_path[MAX_DRIVES];  /* What directory we started out in */

int msgs[MAX_MSG];
HASH_SIZE *users=NULL;

int quiet,
    num_users,
    num_msg;

void Hash_User_File(char *username);
HASH_SIZE hash_code(char *s);
int MsgCRC(int msgnum);

#ifndef __TURBOC__
#undef EmbedCopyright
#define EmbedCopyright()
#endif

void pascal Scan_More(void);

int main(int argc,char *argv[])
{
  FILE *tosslog;

  FFIND *ff;
  struct _msg msg;
  struct _area area;

  int areasfile;

  int msgfile,
      datfile;

  char etoss[PATHLEN],
       temp[PATHLEN];

  char *names_to_scan[MAX_SCANAREA];

  int do_build,
      full_compile,
      partial_compile,
      old_times,
      ret,
      do_echo,
      do_matrix,
      do_local,
      do_it,
      num_areas,
      num_nareas,
      num_names,
      high_msg,
      no_del,
      nmsg,
      x,z;

  int areas_to_scan[MAX_SCANAREA],
      areas_not_to_scan[MAX_SCANAREA];

  long start,
       end;
     
  HASH_SIZE y;

  Hello("SCANBLD","Mail database update utility",VERSION,"1990");

  if (argc < 3 || eqstri(argv[1],"?"))
  {
    printf("Error in command-line!  Format:\n\n");

    printf("SCANBLD <user_file> <adat_file> [area_type...] [@tosslog] [[-x]...]\n\n");

    printf("<user_file> is the path and filename of USER.BBS.\n");
    printf("<adat_file> is the path and filename of AREA.DAT.\n\n");

    printf("[area_type] is optional, and you may specify more than one type to scan.\n");
    printf("[area_type] may be one of the following keywords:\n\n");

    printf("   All - The default.  Scan all areas.\n");
    printf("  Echo - Scan only EchoMail areas.\n");
    printf(" Local - Scan only local message areas.\n");
    printf("Matrix - Scan only Matrix/NetMail areas.\n\n");

    printf("In addition to specifying certain area types, you can also specify individual\n");
    printf("areas to scan.  To do this, place the number of the area on the command line,\n");
    printf("just as you would any of the above types.  To EXCLUDE a specific area from\n");
    printf("being scanned, do the same as above, except prefix the area number with an\n");
    printf("exclamation mark.  (ie. Use `SCANBLD USER.BBS AREAS.DAT ECHO !45' to scan\n");
    printf("all EchoMail areas except for area #45.)  These options can be combined with\n");

    Scan_More();

    printf("any of the others, and in any order.\n\n");

    printf("If a tosslog is specified, then SCANBLD will read the area names given in\n");
    printf("the log, and compare them to the echo tag of each area in AREA.DAT, as given\n");
    printf("by the `MsgName' token in AREAS.CTL, and then scan only those areas.  For\n");
    printf("example, use `SCANBLD USER.BBS AREA.DAT @ECHOTOSS.LOG LOCAL' to scan all\n");
    printf("areas listed in the file `ECHOTOSS.LOG', in addition to any local areas.\n\n");

    printf("You can also specify any of the following switches to change SCANBLD's\n");
    printf("operation:\n\n");

    printf("  -C  - Compile ALL areas.  By default, SCANBLD will scan each area first, to\n");
    printf("        determine whether or not it should skip, update, or compile each\n");
    printf("        area.  Specifying -C forces SCANBLD to compile ALL areas.\n");
    printf("  -Q  - Forces quiet operation.  SCANBLD will display a `#' for each area\n");
    printf("        processed, instead of the usual area statistics.\n");
    printf("  -ND - Tells SCANBLD not to delete the @tosslog file after processing.\n");

    exit(1);
  }

  orig_disk=-1;

  *etoss='\0';
  do_echo=do_matrix=do_local=full_compile=quiet=no_del=FALSE;
  num_areas=num_nareas=num_names=0;

  for (x=0;x < MAX_DRIVES;x++)
    orig_path[x]=NULL;

  do_it=FALSE;

  for (x=3;x < argc;x++)
  {
    if (eqstri(argv[x],"all"))
      do_echo=do_matrix=do_local=do_it=TRUE;
    else if (eqstri(argv[x],"echo"))
      do_echo=do_it=TRUE;
    else if (eqstri(argv[x],"local"))
      do_local=do_it=TRUE;
    else if (eqstri(argv[x],"matrix"))
      do_matrix=do_it=TRUE;
    else if (eqstri(argv[x],"/c") || eqstri(argv[x],"-c"))
      full_compile=TRUE;
    else if (eqstri(argv[x],"/q") || eqstri(argv[x],"-q"))
      quiet=TRUE;
    else if (eqstri(argv[x],"/nd") || eqstri(argv[x],"-nd"))
      no_del=TRUE;
    else if (*argv[x]=='!')
    {
      if (num_nareas > MAX_SCANAREA)
        printf("Error!  Too many areas listed for exclusion from scan.\n");
      else areas_not_to_scan[num_nareas++]=Set_Area_Name(argv[x]+1);
    }
    else if (isalnum(*argv[x]))
    {
      if (num_areas >= MAX_SCANAREA)
        printf("Error!  Too many areas listed for inclusion in scan.\n");
      else
      {
        areas_to_scan[num_areas++]=Set_Area_Name(argv[x]);
        do_it=TRUE;
      }
    }
    else if (*argv[x]=='@')
    {
      do_it=TRUE;

      if ((tosslog=shfopen(strcpy(etoss,argv[x]+1),"r",O_RDONLY))==NULL)
      {
        printf("Error opening `%s' for read!\n",etoss);
        continue;
      }

      while (fgets(temp,PATHLEN,tosslog))
      {
        if (*temp && temp[strlen(temp)-1]=='\n')
          temp[strlen(temp)-1]='\0';

        if (num_names >= MAX_SCANAREA)
        {
          printf("Error!  Too many areas listed in %s.\n",etoss);
          break;
        }
        
        if (*temp &&
            (names_to_scan[num_names]=malloc(strlen(temp)+1)) != NULL)
          strcpy(names_to_scan[num_names++],temp);
      }

      fclose(tosslog);
    }
    else
    {
      printf("\aInvalid command-line parameter: `%s'\n",argv[x]);
      exit(1);
    }
  }

  if (*etoss && !no_del)
    unlink(etoss);

  if (do_it==FALSE)
    do_local=do_echo=do_matrix=TRUE;


  Hash_User_File(argv[1]);

  nmsg=0;
  start=time(NULL);

  if ((areasfile=shopen(argv[2],O_RDONLY | O_BINARY))==-1)
  {
    printf("Error opening `%s' for read!\n",argv[2]);
    exit(1);
  }

  while (read(areasfile,(char *)&area,sizeof(struct _area))==
           sizeof(struct _area))
  {
    if (area.id != AREA_id)
    {
      printf("\aInvalid or corrupted area data file: `%s'!\n",argv[2]);
      exit(1);
    }

    do_it=TRUE;

    if ((area.attrib[0] & ECHOMAIL) && !do_echo)
      do_it=FALSE;

    if ((area.attrib[0] & SYSMAIL) && !do_matrix)
      do_it=FALSE;

    if (!(area.attrib[0] & (SYSMAIL | ECHOMAIL)) && !do_local)
      do_it=FALSE;

    if (do_it==FALSE)
    {
      for (x=0;x < num_areas;x++)
        if (areas_to_scan[x]==area.areano)
        {
          do_it=TRUE;
          break;
        }
    }

    if (do_it==FALSE)
    {
      for (x=0;x < num_names;x++)
         if (eqstri(names_to_scan[x],area.msgname))
        {
          do_it=TRUE;
          break;
        }
    }

    if (do_it==TRUE)
    {
      for (x=0;x < num_nareas;x++)
        if (areas_not_to_scan[x]==area.areano)
        {
          do_it=FALSE;
          break;
        }
    }

    if (do_it && *area.msgpath)
    {
      strcpy(temp,area.msgpath);

      if (!direxist(temp) || Save_Dir(&orig_disk,orig_path,temp)==-1)
        printf("Can't find path `%s'!\n",temp);
      else
      {
        if (quiet)
          printf("#");
        else
        {
          if (isatty(fileno(stdout)))
            printf("Area %-40s - Scanning",temp);
          else printf("Area %-40s -",temp);
        }

        ff=FindOpen("*.MSG",0);
        num_msg=high_msg=ret=0;

        if (ff)
        {
          for (ret=0;ret==0 && num_msg < MAX_MSG;ret=FindNext(ff))
          {
            /* Skip this message if invalid */
            if ((msgs[num_msg]=atoi(ff->ff_name)) != 0)
            {
              if (msgs[num_msg] > high_msg)
                high_msg=msgs[num_msg];

              num_msg++;
            }
          }
        }

        FindClose(ff);

        if (! quiet)
          if (isatty(fileno(stdout)))
            printf("\x08\x08\x08\x08\x08\x08\x08\x08(%4d)  \x08- ",num_msg);
          else printf(" (%4d) - ",num_msg);

        if (num_msg==MAX_MSG)
        {
          printf("\nToo many messages in %s (max %d)\n",temp,MAX_MSG);
          continue;
        }

        /* Sort the messages into numerical order! */
        qksort(msgs,num_msg);

        do_build=partial_compile=FALSE;

        if ((datfile=shopen("SCANFILE.DAT",O_RDONLY | O_BINARY))==-1)
          do_build=TRUE;
        else
        {
          read(datfile,(char *)&msg,sizeof(struct _msg));

          if (msg.cost != num_msg || high_msg != msg.times)
          {
            /* If both num_msg and high_msg were incremented the same      *
             * amount -- Probably means that one or more msgs were entered *
             * and nothing else.  That means we can just do a partial      *
             * recompile!                                                  */

            if (num_msg-msg.cost==high_msg-msg.times)
              partial_compile=do_build=TRUE;
            else do_build=TRUE;
          }

          close(datfile);

          if (MsgCRC(msg.times) != msg.attr)
          {
            partial_compile=FALSE;
            do_build=TRUE;
          }
        }

        if (full_compile)       /* Cmd-line switch overrides this */
          partial_compile=FALSE;

        if (partial_compile)
          old_times=msg.times;

        if (do_build || full_compile)
        {
          if (partial_compile)
          {
            if ((datfile=shopen("SCANFILE.DAT",O_RDWR | O_BINARY))==-1)
            {
              printf("Error opening `SCANFILE.DAT' for read/write!\n");
              continue;
            }
          }
          else
          {
            if ((datfile=sopen("SCANFILE.DAT",O_WRONLY | O_CREAT | O_TRUNC |
                              O_BINARY,SH_DENYNONE,S_IREAD | S_IWRITE))==-1)
            {
              printf("Error opening `SCANFILE.DAT' for write!\n");
              continue;
            }
          }

          lseek(datfile,0L,SEEK_SET);

          memset(&msg,'\0',sizeof(struct _msg));
          msg.cost=num_msg;
          msg.times=high_msg;
          msg.attr=MsgCRC(high_msg);

          strcpy(msg.from,"ScanBld");
          strcpy(msg.to,"ScanBld");

          write(datfile,(char *)&msg,sizeof(struct _msg));

          if (!partial_compile)
          {
            if (!quiet)
              printf("Rebuilding");

            x=0;
          }
          else
          {
            lseek(datfile,0L,SEEK_END);

            if (!quiet)
              printf("Updating  ");

            while (x < num_msg && msgs[x] <= old_times)
              x++;
          }

          for (;x < num_msg;x++)
          {
            sprintf(temp,"%d.MSG",msgs[x]);

            if ((msgfile=shopen(temp,O_RDONLY | O_BINARY))==-1)
              printf("Error opening `%s' for read!\n",temp);
            else
            {
              nmsg++;
              read(msgfile,(char *)&msg,sizeof(struct _msg));
              close(msgfile);

              y=hash_code(msg.to);

              for (z=0;z < num_users;z++)
              {
                if (y==users[z])
                {
                  msg.cost=msgs[x];
                  write(datfile,(char *)&msg,sizeof(struct _msg));
                  break;
                }
              }
            }
          }

          close(datfile);

          if (!quiet)
            printf(" - Done\n");
        }
        else if (!quiet)
          printf("Skipping\n");
      }
    }
  }


  Restore_Dir(&orig_disk,orig_path);

  close(areasfile);

  for (x=0;x < num_names;x++)
    free(names_to_scan[x]);

  end=time(NULL)-start;

  if (end)
    printf("\nDone.  Compiled %d messages in %ld seconds (%d.%d msgs/sec).\n",
           nmsg,end,(int)((long)nmsg/end),(int)((long)nmsg*10L/end) % 10);

  if (users)
    free(users);
  
  return 0;
}




#define xcrc(ch,crc) (((crc>>8)&0x00ff) ^ crctab[(crc^ch)&0x00ff])

unsigned short crctab[]={
  0x0000,0xC0C1,0xC181,0x0140,0xC301,0x03C0,0x0280,0xC241,
  0xC601,0x06C0,0x0780,0xC741,0x0500,0xC5C1,0xC481,0x0440,
  0xCC01,0x0CC0,0x0D80,0xCD41,0x0F00,0xCFC1,0xCE81,0x0E40,
  0x0A00,0xCAC1,0xCB81,0x0B40,0xC901,0x09C0,0x0880,0xC841,
  0xD801,0x18C0,0x1980,0xD941,0x1B00,0xDBC1,0xDA81,0x1A40,
  0x1E00,0xDEC1,0xDF81,0x1F40,0xDD01,0x1DC0,0x1C80,0xDC41,
  0x1400,0xD4C1,0xD581,0x1540,0xD701,0x17C0,0x1680,0xD641,
  0xD201,0x12C0,0x1380,0xD341,0x1100,0xD1C1,0xD081,0x1040,
  0xF001,0x30C0,0x3180,0xF141,0x3300,0xF3C1,0xF281,0x3240,
  0x3600,0xF6C1,0xF781,0x3740,0xF501,0x35C0,0x3480,0xF441,
  0x3C00,0xFCC1,0xFD81,0x3D40,0xFF01,0x3FC0,0x3E80,0xFE41,
  0xFA01,0x3AC0,0x3B80,0xFB41,0x3900,0xF9C1,0xF881,0x3840,
  0x2800,0xE8C1,0xE981,0x2940,0xEB01,0x2BC0,0x2A80,0xEA41,
  0xEE01,0x2EC0,0x2F80,0xEF41,0x2D00,0xEDC1,0xEC81,0x2C40,
  0xE401,0x24C0,0x2580,0xE541,0x2700,0xE7C1,0xE681,0x2640,
  0x2200,0xE2C1,0xE381,0x2340,0xE101,0x21C0,0x2080,0xE041,
  0xA001,0x60C0,0x6180,0xA141,0x6300,0xA3C1,0xA281,0x6240,
  0x6600,0xA6C1,0xA781,0x6740,0xA501,0x65C0,0x6480,0xA441,
  0x6C00,0xACC1,0xAD81,0x6D40,0xAF01,0x6FC0,0x6E80,0xAE41,
  0xAA01,0x6AC0,0x6B80,0xAB41,0x6900,0xA9C1,0xA881,0x6840,
  0x7800,0xB8C1,0xB981,0x7940,0xBB01,0x7BC0,0x7A80,0xBA41,
  0xBE01,0x7EC0,0x7F80,0xBF41,0x7D00,0xBDC1,0xBC81,0x7C40,
  0xB401,0x74C0,0x7580,0xB541,0x7700,0xB7C1,0xB681,0x7640,
  0x7200,0xB2C1,0xB381,0x7340,0xB101,0x71C0,0x7080,0xB041,
  0x5000,0x90C1,0x9181,0x5140,0x9301,0x53C0,0x5280,0x9241,
  0x9601,0x56C0,0x5780,0x9741,0x5500,0x95C1,0x9481,0x5440,
  0x9C01,0x5CC0,0x5D80,0x9D41,0x5F00,0x9FC1,0x9E81,0x5E40,
  0x5A00,0x9AC1,0x9B81,0x5B40,0x9901,0x59C0,0x5880,0x9841,
  0x8801,0x48C0,0x4980,0x8941,0x4B00,0x8BC1,0x8A81,0x4A40,
  0x4E00,0x8EC1,0x8F81,0x4F40,0x8D01,0x4DC0,0x4C80,0x8C41,
  0x4400,0x84C1,0x8581,0x4540,0x8701,0x47C0,0x4680,0x8641,
  0x8201,0x42C0,0x4380,0x8341,0x4100,0x81C1,0x8081,0x4040
};


int MsgCRC(int msgnum)
{
  struct _msg msg;

  char temp[PATHLEN],
       *s,
       *e;

  int msgfile,
      crc;

  sprintf(temp,"%d.MSG",msgnum);

  if ((msgfile=shopen(temp,O_RDONLY | O_BINARY))==-1)
    return 0;

  if (read(msgfile,(char *)&msg,sizeof(struct _msg)) != sizeof(struct _msg))
  {
    close(msgfile);
    return 0;
  }

  close(msgfile);

  /* So that a changed attribute (or reply) doesn't affect things */
  msg.attr=msg.up=msg.reply=0;

  for (crc=0xffff,s=(char *)&msg,e=s+sizeof(struct _msg);s < e;s++)
    crc=xcrc(*s,crc);

  return (xcrc(0,xcrc(0,crc)));
}


HASH_SIZE hash_code(char *s)
{
  HASH_SIZE hash;
  int ch;

  for (hash=0xffff;*s;s++)
  {
    ch=tolower(*s);
    hash=xcrc(ch,hash);
  }
  
  return (xcrc(0,xcrc(0,hash)));
}


void Hash_User_File(char *username)
{
  int userfile,
      num_recs;
    
  struct _usr usr;
  
  if ((userfile=shopen(username,O_RDONLY | O_BINARY))==-1)
  {
    printf("Error opening `%s' for read!  Aborting...\n",username);
    exit(1);
  }

  num_recs=(int)(fsize(username)/(long)sizeof(struct _usr));
  
  if ((users=malloc((num_recs+1)*sizeof(HASH_SIZE)))==NULL)
  {
    printf("Not enough memory to hash USER.BBS!  Aborting...\n");
    exit(1);
  }

  if (!quiet)
    printf("Hashing USER.BBS ");

  for (num_users=0;
       read(userfile,(char *)&usr,sizeof(struct _usr))==sizeof(struct _usr) &&
       num_users < num_recs;)
    users[num_users++]=hash_code(usr.name);

  close(userfile);

  if (num_users > num_recs)
  {
    printf("- Hash memory allocation error.  GRONK.\n");
    exit(1);
  }

  if (!quiet)
    printf("(%d users) - Done\n\n",num_users);
}


void pascal Scan_More(void)
{
  printf("[Press Enter] ");
  fgetchar();
  printf("\x1b[A\r              \r");
}


